use std::sync::Arc;

use axum::{
    extract::State,
    http::{HeaderMap, HeaderValue},
    response::{ErrorResponse, Html, IntoResponse, Redirect},
    routing::{get, post, IntoMakeService},
    Json, Router,
};
use serde::Serialize;
use tokio::sync::RwLock;

use crate::{
    task::{collect_local_address, MY_ADDRESS},
    types::{AppState, PageContext},
    web::templates::load_templates,
};

enum Response<T> {
    Html(Html<String>),
    Json(Json<T>),
}

impl<T> IntoResponse for Response<T>
where
    T: Serialize,
{
    fn into_response(self) -> axum::response::Response {
        match self {
            Response::Html(html) => html.into_response(),
            Response::Json(json) => json.into_response(),
        }
    }
}

#[derive(Serialize, Default)]
struct AddressResult {
    address: String,
}

async fn root(
    State(state): State<Arc<RwLock<AppState>>>,
    headers: HeaderMap,
) -> axum::response::Result<Response<AddressResult>> {
    let address_guard = MY_ADDRESS.lock().await;
    let address = address_guard.clone();
    drop(address_guard);

    let accept: Vec<&str> = headers
        .get("Accept")
        .map(|h| h.to_str().unwrap())
        .unwrap_or("text/html")
        .split(",")
        .collect();

    tracing::info!(?accept, "accept");
    if accept.contains(&"text/html") {
        let state = state.read().await;

        let ctx = match tera::Context::from_serialize(&PageContext { address }) {
            Ok(ctx) => ctx,
            Err(err) => return Err(ErrorResponse::from(format!("{err}"))),
        };
        match state.templates.render("index.html", &ctx) {
            Ok(result) => Ok(Response::Html(Html::from(result))),
            Err(err) => Err(ErrorResponse::from(format!("{err}"))),
        }
    } else if accept.contains(&"application/json") {
        Ok(Response::Json(Json(AddressResult { address })))
    } else {
        unreachable!()
    }
}

#[derive(Serialize, Default)]
struct ReloadResult {
    #[serde(skip_serializing_if = "Option::is_none")]
    error: Option<String>,
}

async fn reload(
    State(state): State<Arc<RwLock<AppState>>>,
) -> Result<Redirect, Json<ReloadResult>> {
    let mut state = state.write_owned().await;
    if let Err(err) = state.templates.full_reload() {
        return Err(Json(ReloadResult {
            error: Some(err.to_string()),
        }));
    }
    drop(state);
    if let Err(err) = collect_local_address().await {
        return Err(Json(ReloadResult {
            error: Some(err.to_string()),
        }));
    }

    Ok(Redirect::to("/"))
}

pub fn make_service() -> Result<IntoMakeService<Router<()>>, Box<dyn std::error::Error>> {
    let templates = load_templates()?;

    let router = Router::new();

    #[cfg(feature = "serve-static")]
    let router = router.route_service("/static/*path", staticfiles::StaticFiles::strip("/static/"));

    Ok(router
        .route("/", get(root))
        .route("/reload", post(reload))
        .with_state(Arc::new(RwLock::new(crate::types::AppState {
            templates: templates,
        })))
        .into_make_service())
}
